"""Tests for the `cli` module."""

from __future__ import annotations

import threading
import time
from typing import TYPE_CHECKING

import pytest

from aria2p import debug
from aria2p.cli.commands import top
from aria2p.cli.commands.add_magnet import add_magnets
from aria2p.cli.commands.add_metalink import add_metalinks
from aria2p.cli.commands.add_torrent import add_torrents
from aria2p.cli.commands.call import call
from aria2p.cli.commands.listen import listen
from aria2p.cli.commands.pause import pause
from aria2p.cli.commands.purge import purge
from aria2p.cli.commands.remove import remove
from aria2p.cli.commands.resume import resume
from aria2p.cli.commands.show import show
from aria2p.cli.main import main
from aria2p.cli.parser import get_parser
from tests import BUNSENLABS_MAGNET, TESTS_DATA_DIR
from tests.conftest import Aria2Server

if TYPE_CHECKING:
    from pathlib import Path


def out_lines(cs: pytest.CaptureFixture) -> str:
    return cs.readouterr().out.split("\n")


def err_lines(cs: pytest.CaptureFixture) -> str:
    return cs.readouterr().err.split("\n")


def first_out_line(cs: pytest.CaptureFixture) -> str:
    return out_lines(cs)[0]


def first_err_line(cs: pytest.CaptureFixture) -> str:
    return err_lines(cs)[0]


def test_show_help(capsys: pytest.CaptureFixture) -> None:
    """Show help.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["-h"])
    captured = capsys.readouterr()
    assert "aria2p" in captured.out


def test_main_returns_2_when_no_remote_running(port: int) -> None:
    assert main([f"--port={port}"]) == 2


def test_parser_error_when_gids_and_all_option(capsys: pytest.CaptureFixture) -> None:
    with pytest.raises(SystemExit) as e:  # noqa: PT012
        main(["pause", "-a", "0000000000000001"])
        assert e.value.code == 2
    lines = err_lines(capsys)
    assert lines[0].startswith("usage: aria2p pause")
    assert lines[1].endswith("-a/--all: not allowed with arguments gids")


def test_parser_error_when_no_gid_and_no_all_option(capsys: pytest.CaptureFixture) -> None:
    def assert_func(command: str, alias: str) -> None:
        with pytest.raises(SystemExit) as e:  # noqa: PT012
            main([alias])
            assert e.value.code == 2
        lines = err_lines(capsys)
        assert lines[0].startswith("usage: aria2p " + command)
        assert lines[1].endswith("the following arguments are required: gids or --all")

    for command, aliases in [
        ("pause", ["stop"]),
        ("remove", ["rm", "del", "delete"]),
        ("resume", ["start"]),
    ]:
        assert_func(command, command)
        for alias in aliases:
            assert_func(command, alias)


def test_no_interface_deps_print_error(
    server: Aria2Server,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture,
) -> None:
    monkeypatch.setattr(top, "Interface", None)
    main(["-p", str(server.port)])
    line = first_err_line(capsys)
    assert "aria2p[tui]" in line


def test_main_show_subcommand(server: Aria2Server, capsys: pytest.CaptureFixture) -> None:
    main(["-p", str(server.port), "show"])
    first_line = first_out_line(capsys)
    for word in ("GID", "STATUS", "PROGRESS", "DOWN_SPEED", "UP_SPEED", "ETA", "NAME"):
        assert word in first_line


def test_errors_and_print_message(server: Aria2Server, capsys: pytest.CaptureFixture) -> None:
    assert main(["-p", str(server.port), "call", "tellstatus", "-P", "invalid gid"]) > 0
    assert capsys.readouterr().err == "Invalid GID invalid gid\n"


def test_show_subcommand(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
        assert show(server.api) == 0
        assert len(capsys.readouterr().out.rstrip("\n").split("\n")) == 2


def test_call_subcommand(server: Aria2Server, capsys: pytest.CaptureFixture) -> None:
    assert call(server.api, "wrongMethod", []) == 1
    assert (
        capsys.readouterr().err == "aria2p: call: Unknown method wrongMethod.\n"
        "  Run 'aria2p call listmethods' to list the available methods.\n"
    )


def test_call_subcommand_with_json_params(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
        assert call(server.api, "tellstatus", '["0000000000000001"]') == 0


def test_call_subcommand_with_no_params(server: Aria2Server) -> None:
    assert call(server.api, "listmethods", []) == 0


def test_add_magnet_subcommand(server: Aria2Server) -> None:
    assert add_magnets(server.api, [BUNSENLABS_MAGNET]) == 0


def test_add_torrent_subcommand(server: Aria2Server) -> None:
    assert add_torrents(server.api, [TESTS_DATA_DIR / "bunsenlabs-helium-4.iso.torrent"]) == 0


def test_add_metalink_subcommand(server: Aria2Server) -> None:
    assert add_metalinks(server.api, [TESTS_DATA_DIR / "debian.metalink"]) == 0


def test_pause_subcommand(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="1-dl.txt") as server:
        retcode = pause(server.api, ["0000000000000001"])
    if "cannot be paused now" in capsys.readouterr().err:
        pytest.xfail("Cannot pause download (sporadic error)")
    assert retcode == 0


def test_pause_subcommand_already_paused(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        assert pause(server.api, ["0000000000000001", "0000000000000002"]) == 1
        assert (
            capsys.readouterr().err
            == "GID#0000000000000001 cannot be paused now\nGID#0000000000000002 cannot be paused now\n"
        )


def test_pause_subcommand_one_paused(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="one-active-one-paused.txt") as server:
        assert pause(server.api, ["0000000000000001", "0000000000000002"]) == 1
        assert "GID#0000000000000002 cannot be paused now" in capsys.readouterr().err


def test_pause_all_subcommand(server: Aria2Server) -> None:
    assert pause(server.api, do_all=True) == 0


def test_pause_all_subcommand_doesnt_fail_with_already_paused_downloads(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:
        assert pause(server.api, do_all=True) == 0


def test_resume_subcommand(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
        assert resume(server.api, ["0000000000000001"]) == 0


def test_resume_subcommand_already_unpaused(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="2-dls.txt") as server:
        assert resume(server.api, ["0000000000000001", "0000000000000002"]) == 1
        assert (
            capsys.readouterr().err
            == "GID#0000000000000001 cannot be unpaused now\nGID#0000000000000002 cannot be unpaused now\n"
        )


def test_resume_subcommand_one_unpaused(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="one-active-one-paused.txt") as server:
        assert resume(server.api, ["0000000000000001", "0000000000000002"]) == 1
        err = capsys.readouterr().err
        if "is not found" in err:
            pytest.xfail("Download cannot be found (sporadic error)")
        assert err == "GID#0000000000000001 cannot be unpaused now\n"


def test_resume_all_subcommand(server: Aria2Server) -> None:
    assert resume(server.api, do_all=True) == 0


def test_resume_all_subcommand_doesnt_fail_with_already_active_downloads(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="2-dls.txt") as server:
        assert resume(server.api, do_all=True) == 0


def test_remove_subcommand(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
        assert remove(server.api, ["0000000000000001"]) == 0


def test_remove_subcommand_one_failure(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
        assert remove(server.api, ["0000000000000001", "0000000000000002"]) == 1
        assert capsys.readouterr().err == "GID 0000000000000002 is not found\n"


def test_remove_all_subcommand(server: Aria2Server) -> None:
    assert remove(server.api, do_all=True) == 0


def test_purge_subcommand(tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="very-small-download.txt") as server:
        assert purge(server.api) == 0


def test_listen_subcommand(tmp_path: Path, port: int, capsys: pytest.CaptureFixture) -> None:
    with Aria2Server(tmp_path, port, session="2-dls-paused.txt") as server:

        def thread_target() -> None:
            time.sleep(2)
            server.api.resume_all()
            time.sleep(3)
            server.api.stop_listening()

        thread = threading.Thread(target=thread_target)
        thread.start()
        listen(server.api, callbacks_module=TESTS_DATA_DIR / "callbacks.py", event_types=["start"])
    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out == "started 0000000000000001\nstarted 0000000000000002\n"


@pytest.mark.parametrize("command", ["add", "add-magnet", "add-torrent", "add-metalink"])
def test_parse_valid_options(command: str) -> None:
    parser = get_parser()
    opts = parser.parse_args([command, "/some/file/on/the/disk", "-o", "opt1=val;opt2=val2, val3=val4"])
    assert opts.options == {"opt1": "val", "opt2": "val2, val3=val4"}


@pytest.mark.parametrize("command", ["add", "add-magnet", "add-torrent", "add-metalink"])
def test_parse_invalid_options(command: str, capsys: pytest.CaptureFixture) -> None:
    parser = get_parser()
    with pytest.raises(SystemExit):
        parser.parse_args([command, "http://example.com", "-o", "opt1"])
    assert "Options strings must follow this format" in capsys.readouterr().err


@pytest.mark.parametrize(
    ("command", "option"),
    [("add", "uris"), ("add-magnet", "uris"), ("add-torrent", "torrent_files"), ("add-metalink", "metalink_files")],
)
def test_error_when_missing_arg(command: str, option: str, capsys: pytest.CaptureFixture) -> None:
    with pytest.raises(SystemExit):
        main([command])
    assert option in capsys.readouterr().err


def test_show_version(capsys: pytest.CaptureFixture) -> None:
    """Show version.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["-V"])
    captured = capsys.readouterr()
    assert debug.get_version() in captured.out


def test_show_debug_info(capsys: pytest.CaptureFixture) -> None:
    """Show debug information.

    Parameters:
        capsys: Pytest fixture to capture output.
    """
    with pytest.raises(SystemExit):
        main(["--debug-info"])
    captured = capsys.readouterr().out.lower()
    assert "python" in captured
    assert "system" in captured
    assert "environment" in captured
    assert "packages" in captured


@pytest.mark.parametrize("command", ["autoclear", "autopurge", "autoremove"])
def test_cli_autoclear_commands(command: str, tmp_path: Path, port: int) -> None:
    with Aria2Server(tmp_path, port, session="very-small-download.txt"):
        assert main([command]) == 0



"""Command to purge downloads."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aria2p.api import API


def purge(api: API) -> int:
    """Purge subcommand.

    Parameters:
        api: The API instance to use.

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if api.autopurge():
        return 0
    return 1



"""Command to add metalinks."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from aria2p.utils import read_lines

if TYPE_CHECKING:
    from pathlib import Path

    from aria2p.api import API


def add_metalinks(
    api: API,
    metalink_files: list[str | Path] | None = None,
    from_file: str | None = None,
    options: dict | None = None,
    position: int | None = None,
) -> int:
    """Add metalink subcommand.

    Parameters:
        api: The API instance to use.
        metalink_files: The paths to the metalink files.
        from_file: Path to the file to metalink files paths from.
        options: String of aria2c options to add to download.
        position: Position to add new download in the queue.

    Returns:
        int: 0 if OK else 1.
    """
    ok = True

    if not metalink_files:
        metalink_files = []

    if from_file:
        try:
            metalink_files.extend(read_lines(from_file))
        except OSError:
            print(f"Cannot open file: {from_file}", file=sys.stderr)
            ok = False

    for metalink_file in metalink_files:
        new_downloads = api.add_metalink(metalink_file, options=options, position=position)
        for download in new_downloads:
            print(f"Created download {download.gid}")

    return 0 if ok else 1



"""Tests for the `downloads` module."""

from __future__ import annotations

import datetime
from datetime import timezone
from pathlib import Path

import pytest

from aria2p import API, BitTorrent, ClientException, Download, File, Options
from tests.conftest import Aria2Server


class TestBitTorrentClass:
    def setup_method(self) -> None:
        self.bittorrent = BitTorrent(
            {"announceList": [], "comment": "", "creationDate": 10, "mode": "single", "info": {"name": ""}},
        )

    def test_init_method(self) -> None:
        assert BitTorrent(self.bittorrent._struct)

    def test_str_method(self) -> None:
        assert self.bittorrent.info
        assert str(self.bittorrent) == self.bittorrent.info["name"]

    def test_announce_list_property(self) -> None:
        assert self.bittorrent.announce_list == []

    def test_comment_property(self) -> None:
        assert self.bittorrent.comment == ""

    def test_creation_date_property(self) -> None:
        assert self.bittorrent.creation_date == datetime.datetime.fromtimestamp(10, tz=timezone.utc)

    def test_mode_property(self) -> None:
        assert self.bittorrent.mode == "single"

    def test_info_property(self) -> None:
        assert self.bittorrent.info == {"name": ""}


class TestDownloadClass:
    def setup_method(self) -> None:
        self.bitfield = (
            "9e7ef5fbdefdffffdfffffffffffffdf7fff5dfefff7feffefffdff7fffffffbfeffbffddf7de9f7eefffffffeffe77bb"
            "f8fdbdcffef7fffffbefad7ffffffbf55bff7edfedfeff7ffff7ffffff3ffff7d3ffbfffddddffe7ffffffffffdedf7fd"
            "fef62fbdfffffbffbffdcfaffffafebeff3ebff7d5fffbbb2bafef77ffaffe7d37fbffffffb6dfffffffedfebbecbbffe"
            "bdefffffffff977f7ffdffee7fffffffdfeb3f67dddffeedfffffbfffffdaffbfeedfadef6dfd9d2df9fb7f689ffcff3f"
            "fbfebbfdbd7fcddfa77dfddffdfe327fdcbf77fad87ffff6ff7ffebfefdfbbffdefdafeefed7fefffe7ffad9ffdcffefc"
            "ffbbf3c07ef7fc0"
        )
        self.download = Download(
            API(),
            {
                "bitfield": self.bitfield,
                "bittorrent": {},
                "completedLength": "889592532",
                "connections": "44",
                "dir": "/home/pawamoy/Downloads/torrents/tmp",
                "downloadSpeed": "745509",
                "files": [
                    {
                        "completedLength": "58132",
                        "index": "1",
                        "length": "58132",
                        "path": "/home/pawamoy/Downloads/torrents/tmp/dl/logo.jpg",
                        "selected": "true",
                        "uris": [],
                    },
                    {
                        "completedLength": "885864384",
                        "index": "2",
                        "length": "1045247936",
                        "path": "/home/pawamoy/Downloads/torrents/tmp/dl/video.mp4",
                        "selected": "true",
                        "uris": [],
                    },
                ],
                "following": "a89a9c5ac990e6ef",
                "gid": "0a6635c602761000",
                "infoHash": "4f1da018803b65f61ed76612af9ad00d4373a771",
                "numPieces": "1994",
                "numSeeders": "12",
                "pieceLength": "524288",
                "seeder": "false",
                "status": "active",
                "totalLength": "1045306068",
                "uploadLength": "97207027",
                "uploadSpeed": "11032",
            },
        )

    def test_eq_method(self) -> None:
        assert self.download == Download(API(), {"gid": "0a6635c602761000"})

    def test_init_method(self) -> None:
        assert Download(API(), {})

    def test_str_method(self) -> None:
        assert str(self.download) == self.download.name

    def test_update_method(self, tmp_path: Path, port: int) -> None:
        with Aria2Server(tmp_path, port, session="1-dl-paused.txt") as server:
            download = server.api.get_download("0000000000000001")
            download.update()

    def test_belongs_to(self, server: Aria2Server) -> None:
        self.download.api = server.api
        assert self.download.belongs_to is None

    def test_belongs_to_id(self) -> None:
        assert self.download.belongs_to_id is None

    def test_bitfield(self) -> None:
        assert self.download.bitfield == self.bitfield

    def test_bittorrent(self) -> None:
        assert self.download.bittorrent

        # assert value was cached
        self.download._struct["bittorrent"] = {"mode": "single"}
        assert self.download.bittorrent.mode is None

        # assert is None when "bittorrent" key is not in struct
        del self.download._struct["bittorrent"]
        self.download._bittorrent = None
        assert self.download.bittorrent is None

    def test_completed_length(self) -> None:
        assert self.download.completed_length == 889_592_532

    def test_completed_length_string(self) -> None:
        assert self.download.completed_length_string() == "848.38 MiB"
        assert self.download.completed_length_string(human_readable=False) == "889592532 B"

    def test_connections(self) -> None:
        assert self.download.connections == 44

    def test_dir(self) -> None:
        assert self.download.dir == Path("/home/pawamoy/Downloads/torrents/tmp")

    def test_download_speed(self) -> None:
        assert self.download.download_speed == 745_509

    def test_download_speed_string(self) -> None:
        assert self.download.download_speed_string() == "728.04 KiB/s"
        assert self.download.download_speed_string(human_readable=False) == "745509 B/s"

    def test_error_code(self) -> None:
        assert self.download.error_code is None

    def test_error_message(self) -> None:
        assert self.download.error_message is None

    def test_eta(self) -> None:
        assert isinstance(self.download.eta, datetime.timedelta)
        self.download._struct["downloadSpeed"] = "0"
        assert self.download.eta == datetime.timedelta.max

    def test_eta_string(self) -> None:
        assert self.download.eta_string() == "3m28s"
        self.download._struct["downloadSpeed"] = "0"
        assert self.download.eta_string() == "-"

    def test_files(self) -> None:
        assert len(self.download.files) == 2

        # assert value was cached
        self.download._struct["files"] = []
        assert len(self.download.files) == 2

    def test_followed_by(self, server: Aria2Server) -> None:
        self.download.api = server.api
        assert self.download.followed_by == []

    def test_followed_by_ids(self, server: Aria2Server) -> None:
        self.download.api = server.api
        assert self.download.followed_by_ids == []

    def test_following(self, server: Aria2Server) -> None:
        self.download.api = server.api
        assert self.download.following is None

    def test_following_id(self) -> None:
        assert self.download.following_id == "a89a9c5ac990e6ef"

    def test_gid(self) -> None:
        assert self.download.gid == "0a6635c602761000"

    def test_has_failed(self) -> None:
        assert self.download.has_failed is False

    def test_info_hash(self) -> None:
        assert self.download.info_hash == "4f1da018803b65f61ed76612af9ad00d4373a771"

    def test_is_active(self) -> None:
        assert self.download.is_active is True

    def test_is_complete(self) -> None:
        assert self.download.is_complete is False

    def test_is_paused(self) -> None:
        assert self.download.is_paused is False

    def test_is_removed(self) -> None:
        assert self.download.is_removed is False

    def test_is_waiting(self) -> None:
        assert self.download.is_waiting is False

    def test_move(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.move(2)

    def test_move_down(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.move_down(1)

    def test_move_to(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.move_to(10)

    def test_move_to_bottom(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.move_to_bottom()

    def test_move_to_top(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.move_to_top()

    def test_move_up(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.move_up()

    def test_name(self) -> None:
        assert self.download.name == "dl"

    def test_name_bittorrent(self) -> None:
        # TODO
        pass

    def test_name_metadata(self) -> None:
        # TODO
        pass

    def test_name_filepath(self) -> None:
        # TODO
        pass

    def test_name_uri(self) -> None:
        # TODO
        pass

    def test_num_pieces(self) -> None:
        assert self.download.num_pieces == 1994

    def test_num_seeders(self) -> None:
        assert self.download.num_seeders == 12

    def test_options(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            print(self.download.options)  # noqa: T201

    def test_options2(self) -> None:
        witness = []

        def mocked() -> None:
            witness.append(0)
            self.download._options = True  # type: ignore[assignment]

        self.download.update_options = mocked  # type: ignore[method-assign]
        assert self.download.options is True
        assert witness == [0]
        assert self.download.options is True
        assert witness == [0]

    def test_set_options(self) -> None:
        options = Options(self.download.api, {})
        self.download.options = options
        assert self.download.options is options

    def test_pause(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.pause()

    def test_piece_length(self) -> None:
        assert self.download.piece_length == 524_288

    def test_piece_length_string(self) -> None:
        assert self.download.piece_length_string() == "512.00 KiB"
        assert self.download.piece_length_string(human_readable=False) == "524288 B"

    def test_progress(self) -> None:
        assert self.download.progress == self.download.completed_length / self.download.total_length * 100
        self.download._struct["totalLength"] = 0
        assert self.download.progress == 0.0

    def test_progress_string(self) -> None:
        progress = self.download.completed_length / self.download.total_length * 100
        assert self.download.progress_string() == f"{progress:.2f}%"
        self.download._struct["totalLength"] = 0
        assert self.download.progress_string() == "0.00%"

    def test_remove(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.remove()

    def test_resume(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.resume()

    def test_seeder(self) -> None:
        assert self.download.seeder is False

    def test_status(self) -> None:
        assert self.download.status == "active"

    def test_total_length(self) -> None:
        assert self.download.total_length == 1_045_306_068

    def test_total_length_string(self) -> None:
        assert self.download.total_length_string() == "996.88 MiB"
        assert self.download.total_length_string(human_readable=False) == "1045306068 B"

    def test_update(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.update()

    def test_update_options(self, server: Aria2Server) -> None:
        self.download.api = server.api
        with pytest.raises(ClientException):
            self.download.update_options()

    def test_upload_length(self) -> None:
        assert self.download.upload_length == 97_207_027

    def test_upload_length_string(self) -> None:
        assert self.download.upload_length_string() == "92.70 MiB"
        assert self.download.upload_length_string(human_readable=False) == "97207027 B"

    def test_upload_speed(self) -> None:
        assert self.download.upload_speed == 11032

    def test_upload_speed_string(self) -> None:
        assert self.download.upload_speed_string() == "10.77 KiB/s"
        assert self.download.upload_speed_string(human_readable=False) == "11032 B/s"

    def test_verified_length(self) -> None:
        assert self.download.verified_length == 0

    def test_verified_length_string(self) -> None:
        assert self.download.verified_length_string() == "0.00 B"
        assert self.download.verified_length_string(human_readable=False) == "0 B"

    def test_verify_integrity_pending(self) -> None:
        assert self.download.verify_integrity_pending is None

    def test_is_torrent(self) -> None:
        assert self.download.is_torrent
        del self.download._struct["bittorrent"]
        assert not self.download.is_torrent


class TestFileClass:
    def setup_method(self) -> None:
        self.file = File(
            {
                "index": "1",
                "path": "/some/file/path.txt",
                "length": "2097152",
                "completedLength": "2048",
                "selected": "true",
                "uris": [{"uri": "some uri", "status": "used"}, {"uri": "some other uri", "status": "waiting"}],
            },
        )

    def test_eq_method(self) -> None:
        assert self.file == File({"path": "/some/file/path.txt"})

    def test_init_method(self) -> None:
        assert File({})

    def test_str_method(self) -> None:
        assert str(self.file) == str(self.file.path)

    def test_completed_length(self) -> None:
        assert self.file.completed_length == 2048

    def test_completed_length_string(self) -> None:
        assert self.file.completed_length_string() == "2.00 KiB"
        assert self.file.completed_length_string(human_readable=False) == "2048 B"

    def test_index(self) -> None:
        assert self.file.index == 1

    def test_length(self) -> None:
        assert self.file.length == 2_097_152

    def test_length_string(self) -> None:
        assert self.file.length_string() == "2.00 MiB"
        assert self.file.length_string(human_readable=False) == "2097152 B"

    def test_path(self) -> None:
        assert self.file.path == Path("/some/file/path.txt")

    def test_selected(self) -> None:
        assert self.file.selected is True

    def test_uris(self) -> None:
        assert self.file.uris == [{"uri": "some uri", "status": "used"}, {"uri": "some other uri", "status": "waiting"}]
